import java.util.ArrayList;
import java.util.Iterator;
import java.util.Hashtable;
import java.util.HashSet;
import java.util.Collection;
import java.util.StringTokenizer;

public class Stack
{
      // the actual known size of the stack
      private int size;
      
      // the units that could be in the stack, by player
      private Collection[] unitlist;

      private String name;      

      public Stack(String name)
      {
	 unitlist = new Collection[Titan.numPlayers+1];
	 for(int i = 0 ; i < unitlist.length ; i++)
	    unitlist[i] = new ArrayList();
	 size = 0;
	 this.name = name;
      }
      
      public void setSize(int newsize)
      {
	 size = newsize;
      }

      public int getSize()
      {
	 return size;
      }

      public String getName()
      {
	 return name;
      }
      
      private Unit getLeastReferencedUnit(int type, int player, Collection ignore)
      {
	 Unit shortest = null; 
         Unit cur = null; 
         Iterator it = unitlist[player].iterator(); 
         while(it.hasNext()) 
         { 
            cur = (Unit)it.next(); 
            if(cur.getType() == type 
	       && (shortest == null  
		   || cur.getNumStacks(player)  
		   < shortest.getNumStacks(player))
	       && !ignore.contains(cur))
	    {
               shortest = cur; 
	    }
         } 
         if(shortest == null) 
         { 
            Titan.out("Not enough units of the given type could exist in the given stack.");
	    return null;
         } 
	 return shortest;
      }
      public void revealUnit(int type, int numReveal)
      {
	 for(int i = 0 ; i < unitlist.length ; i++)
	    revealUnit(type, i, numReveal);
      }
      
      public void revealUnit(int type, int player, int numReveal)
      {
	 Collection ignore = new HashSet();
	 for(int i = 0 ; i < numReveal ; i++)
	 {
	    Unit shortest = getLeastReferencedUnit(type,player,ignore);
	    if(shortest == null)
	       return;
	    
	    shortest.setStack(this,player);
	    ignore.add(shortest);       
	 }

	 if(isFullyDetermined(player))
	    removePossibles(player);
      }

      public boolean isFullyDetermined(int player)
      {
	 int numDetermined = 0;
	 Iterator it = unitlist[player].iterator();
	 while(it.hasNext())
	    if( ((Unit)it.next()).getNumStacks(player) == 1)
	       numDetermined++;
	 return numDetermined == size;
      }

      public void removePossibles(int player)
      {
	 Iterator it = unitlist[player].iterator();
	 Unit u;
	 while (it.hasNext())
	 {
	    u = (Unit)it.next();
	    if(u.getNumStacks(player) != 1)
	    {
	       it.remove();
	       u.setStackNot(this, player);
	    }	       
	 }	    
      }
      
      public void addUnit(Unit u)
      {
	 for(int i = 0 ; i < unitlist.length ; i++) 
	    unitlist[i].add(u);
	 size++;
      }

      public void addUnit(Unit u, int player)
      {
	 unitlist[player].add(u);
      }

      /* note : units are only deleted when the whole stack is known,
         so it doesn't matter which one we delete.  To be safe,
         though, we'll follow the procedure for reveal. */
      public void removeUnit(int type)
      { 
	 for(int i = 0 ; i < unitlist.length ; i++) 
	 {
	    Unit shortest = getLeastReferencedUnit(type,i,new HashSet()); 
	    if(shortest != null) 
	       unitlist[i].remove(shortest);
	 }
	 size--;
      }

      public void removeUnit(Unit u, int p)
      {
	 unitlist[p].remove(u);
      }

      public void copyUnitlistOnto(Stack other)
      {
	 for(int i = 0 ; i < unitlist.length ; i++)
	 {
	    other.unitlist[i].clear();
	    other.unitlist[i].addAll(unitlist[i]);
	    Iterator it = other.unitlist[i].iterator();
	    while(it.hasNext())
	       ((Unit)it.next()).alsoIn(other, i);
	 }
      }

      public void view(int player)
      {
	 Titan.say("Stack : " + name + " : " + size );
	 Iterator it = unitlist[player].iterator();
	 Unit u;
	 while(it.hasNext())
	 {
	    u = (Unit)it.next();
	    Titan.say(u.getName() + (u.getNumStacks(player) == 1 ? "" : " ?"));
	 }
      }

      public void view()
      {
	 view(Titan.numPlayers);
      }

}