import java.util.ArrayList;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Collection;
import java.util.StringTokenizer;

import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.InputStreamReader;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.File;

public class Titan
{
      public static final int MAX_PLAYERS = 6;

      static int numPlayers;
      static String[] playerNames;

      static ArrayList actionLog;

      static Hashtable stacks;
      static Hashtable unitTypes;
      static Hashtable unitNames;

      static PrintWriter logOut;

      static boolean logging = false;

      static void printIntro()
      {
	 say("Titan Helper");
	 say("  2005, GPL; Alan : Michael : Jeff");	    
	 say("");
      }

      static void getPlayers()
      {
	 boolean success = false;
	 say("How many players?");
	 printPrompt();
	 while(!success)
	 {
	    try
	    {
	       numPlayers = Integer.parseInt(readLine()) ;
	       success = true;
	    }
	    catch(NumberFormatException nfe)
	    {
	       success = false;
	    }
	    if(success)
	       success = numPlayers > 0 && numPlayers < MAX_PLAYERS; 

	    if(!success)
	    {
	       out("Please enter a number between 0 and " + MAX_PLAYERS + ".");
	       printPrompt();
	    }
	 }
	 
	 playerNames = new String[numPlayers];
	 numPlayers++;
	 for(int i = 0 ; i < playerNames.length ; i++)
	 {
	    out("Enter the name of player " + i);
	    printPrompt();
	    playerNames[i] = readLine();
	 }
      }

      public static String getNameFromUnitType(int type)
      {
	 return (String)unitNames.get(new Integer(type));
      }

      static void printHelp()
      {
	 out("");
	 out("Valid Commands:");
	 out("---------------");
	 listCommands();
      }

      static void listCommands()
      {
	 out("list (l)");
	 out("  stacks");
	 out("  players");
	 out("  unitTypes");
	 out("  commands");
	 out("  allStacks");
	 out("recruit (r)");
	 out("  <stack> [number revealed] <revealed unit type> <recruited unit>");
	 out("view (v)");
	 out("  <stack> [player]");
         out("  log");
	 out("split (s)");
	 out("  <old stack> <new stack> <new size>");
	 out("delete (d)");
	 out("  <stack> [unit]");
	 out("add (a)");
	 out("  <stack> [unit]");
	 out("  <stack> defaults");
	 out("reveal");
	 out("  <stack> <unit> [player]");
	 out("readlog");
	 out("  <logfile>");
	 out("undo (u)");
	 out("help");
	 out("exit");
      }
      
      public static void printPrompt()
      {
	 System.out.print(" > ");
      }

      static void enlog(Action a)
      {
	 actionLog.add(a);

	 if(logging)
	 {
	    logOut.println(a);
	    logOut.flush();
	 }
      }

      static void implement(Action a)
      {
	 enlog(a);
	 a.implement();
      }

      static boolean processCommand(String cmd) throws ParserException
      {
	 cmd = cmd.toLowerCase();

	 String command;
	 StringTokenizer st = new StringTokenizer(cmd, " ");
	 if(!st.hasMoreTokens())
	 {
	    printHelp();
	    return true;
	 }
	 command = st.nextToken();
	 if(command.equals("help"))
	 {
	    printHelp();
	    return true;
	 }
	 if(command.equals("exit") || command.equals("quit"))
	 {
	    return false;
	 }
	 else if(command.equals("undo") || command.equals("u"))
	 {
	    if(st.hasMoreTokens())
	    {
	       throw new ParserException("Recieved extra stuff at end of \"undo\" command");
	    }
	    doUndo();
	 }
	 else
	 {
	    if(!st.hasMoreTokens())
	       throw new ParserException("Incomplete command");
	    String arg = st.nextToken();
	    if(command.equals("list") || command.equals("l"))
	    {
	       if(st.hasMoreTokens())
		  throw new ParserException("Recieved extra stuff at end of \"list\" command");
	       else if(arg.equals("commands"))
		  listCommands();
	       else if(arg.equals("unittypes"))
		  listUnitTypes();
	       else if(arg.equals("players"))
		  listPlayers();
	       else if(arg.equals("stacks"))
		  listStacks(true);
	       else if(arg.equals("allstacks"))
		  listStacks(false);
	       else
		  throw new ParserException("Invalid argument to list: " + arg);
	    }
	    else if(command.equals("recruit") || command.equals("r"))
	    {
	       String stack = arg;
	       if(!st.hasMoreTokens())
		  throw new ParserException("Recruit needs at least two arguments");
	       String numRevealed = null;
	       String unitTypeRevealed = null;
	       String unitTypeAdded = null;

	       unitTypeAdded = st.nextToken();
	       if(st.hasMoreTokens())
	       {
		  unitTypeRevealed = unitTypeAdded;
		  unitTypeAdded = st.nextToken();
	       }
	       if(st.hasMoreTokens())
	       {
		  numRevealed = unitTypeRevealed;
		  unitTypeRevealed = unitTypeAdded;
		  unitTypeAdded = st.nextToken();
	       }
	       if(st.hasMoreTokens())
		  throw new ParserException("Too many arguments given to the recruit command");

	       if(!isValidStack(stack))
		  throw new ParserException("Stack " + stack + " not found");
	       if(!isValidUnit(unitTypeAdded))
		  throw new ParserException("Unit type "+  unitTypeAdded + " not found");
	       if(unitTypeRevealed != null && !isValidUnit(unitTypeRevealed))
		  throw new ParserException("Unit type " + unitTypeRevealed + " not found");
	       if(numRevealed != null && !isValidNumber(numRevealed,1,3))
		  throw new ParserException("Number to reveal " + numRevealed + " invalid");
	       if(numRevealed == null)
		  numRevealed = "1";
		  
	       /* we validated input, so these won't fail */
	       implement(Action.revealUnit(stack,unitTypeRevealed,numRevealed));
	       implement(Action.add(stack, unitTypeAdded));
	    }
	    else if(command.equals("view") || command.equals("v"))
	    {
	       if(arg.equals("log"))
		  viewLog();
	       else
	       {
		  String player = null;
		  if(st.hasMoreTokens())
		     player = st.nextToken();

		  if(!isValidStack(arg))
		     throw new ParserException("Stack " +  arg + " not found");
		  if(player != null && !isValidPlayer(player))
		     throw new ParserException("Player " +  player + " not found");
		  viewStack(arg  , player);
	       }
	    }
	    else if(command.equals("readlog"))
	       readLog(arg);
	    else if(command.equals("split") || command.equals("s"))
	    {
	       String oldStack = arg;
	       String newStack;
	       String newSize;
	       
	       if(!st.hasMoreTokens())
		  throw new ParserException("split needs more arguments");
	       newStack = st.nextToken();

	       if(!st.hasMoreTokens())
		  throw new ParserException("split needs more arguments");

	       newSize = st.nextToken();

	       if(!isValidStack(oldStack))
		  throw new ParserException("Stack " + oldStack + " not found");
	       if(!isValidStack(newStack))
		  throw new ParserException("Stack " + newStack + " not found");
	       else if ( ((Stack)stacks.get(newStack)).getSize() != 0)
		  throw new ParserException("Stack " + newStack + " already in use");
	       if(!isValidNumber(newSize,2, ((Stack)stacks.get(oldStack)).getSize()-2))
		  throw new ParserException("Can't split a stack of size " + newSize + " off " + oldStack + "(size " +  ((Stack)stacks.get(oldStack)).getSize() + ")");

	       implement(Action.splitStack(oldStack, newStack, newSize));
	    }
	    else if(command.equals("delete") || command.equals("kill") || command.equals("d"))
	    {
	       String unit = null;
	       if(st.hasMoreTokens()) 
                  unit = st.nextToken(); 
	       
	       if(!isValidStack(arg))
		  throw new ParserException("Stack " + arg + " not found");

	       if(unit != null && !isValidUnit(unit))
		  throw new ParserException("Unit type "+  unit + " not found"); 		  

	       implement(Action.delete(arg, unit));
	    }
	    else if(command.equals("add") || command.equals("a"))
	    {
	       String unit = null;
	       if(st.hasMoreTokens()) 
                  unit = st.nextToken(); 

	       /* an invalid stack is ok if there's no unit, because we're creating it */
	       if(unit != null && !isValidStack(arg)) 
                  throw new ParserException("Stack " + arg + " not found"); 
 
               if(unit != null && !unit.equals("defaults") && !isValidUnit(unit))
                  throw new ParserException("Unit type "+  unit + " not found");          

	       if(unit != null && unit.equals("defaults"))
	       {
		  implement(Action.add(arg,"centaur"));
		  implement(Action.add(arg,"centaur"));
		  implement(Action.add(arg,"ogre"));
		  implement(Action.add(arg,"ogre"));
		  implement(Action.add(arg,"gargoyle"));
		  implement(Action.add(arg,"gargoyle"));
		  implement(Action.add(arg,"titan"));
		  implement(Action.add(arg,"angel"));
	       }
	       else
		  implement(Action.add(arg, unit));
	    }
	    else if(command.equals("reveal"))
	    {
	       String stack = arg;
	       String unitType = null;
	       String numRevealed = null;
	       String player = null;

	       if(!st.hasMoreTokens())
		  throw new ParserException("reveal needs more arguments"); 

	       numRevealed = st.nextToken();

	       if(!isValidNumber(numRevealed,1,7))
	       {
		  unitType = numRevealed;
		  numRevealed = null;
	       }
		  
	       if(numRevealed != null)
	       {
		  if(!st.hasMoreTokens())
		     throw new ParserException("reveal needs more arguments");  
		  unitType = st.nextToken();
	       }
	       
	       if(st.hasMoreTokens())
		  player = st.nextToken();

	       if(!isValidStack(stack))
		  throw new ParserException("Stack " + stack + " not found");
	       if(!isValidUnit(unitType))
		  throw new ParserException("Unit type "+  unitType + " not found");
	       if(player != null && !isValidPlayer(player))
		  throw new ParserException("Player "+  player + " not found");
	       
	       if(numRevealed == null)
		  numRevealed = "1";

	       implement(Action.revealUnit(stack,unitType,player,numRevealed));
	    }
	    else
	       throw new ParserException("Incomplete command");
	 }	 
	 return true;
      }

      public static void readLog(String logfile) throws ParserException
      {
	 try
	 {
	    BufferedReader br = new BufferedReader(
	       new InputStreamReader(
		  new FileInputStream(
		     logfile)));
	    String line; 
	    int lnum = 0; 
	    while((line = br.readLine()) !=  null) 
            { 
               if(!line.startsWith("#")) 
               { 
		  processCommand(line);
	       }
	    }
	    br.close();
	 }
	 catch(FileNotFoundException fne)
	 {
	    throw new ParserException("Log file " + logfile + " not found");
	 }
	 catch(IOException ioe)
	 {
	    throw new ParserException("Can't read from file. Sorry");
	 }
      }

      public static String readLine()
      {
	 try
	 {
	    return (new BufferedReader(new InputStreamReader(System.in))).readLine();
	 }
	 catch(IOException ioe)
	 {
	    err("Can't read input");
	    System.exit(27);
	 }
	 return null;
      }

      public static void fillUnitTypesAndStacks(String unitFile, String stackFile)
      {
         unitTypes = new Hashtable();  
	 unitNames = new Hashtable();  

         stacks = new Hashtable();  
	 
	 String line;
	 int lnum = 0;
	 StringTokenizer st;

	 try
	 {
	    BufferedReader br = new BufferedReader(
	       new InputStreamReader(new FileInputStream(unitFile)));

	    while((line = br.readLine()) !=  null)
	    {
	       if(!line.startsWith("#"))
               {
		  if(line.indexOf(" ") != -1)
		  {
		     err("Spaces in unit type " + line);
		     System.exit(7);
		  }
		  line = line.toLowerCase();

		  unitTypes.put(line,new Integer(lnum));
		  unitNames.put(new Integer(lnum), line);
               }
               lnum++;
	    }

	    br.close();
	 }
	 catch(FileNotFoundException f)
	 {
	    err("Can't find the stack listing file " + stackFile);
	    System.exit(55);
	 }
	 catch(IOException f)
	 {
	    err("Can't read from the stack listing file " + stackFile);
	    System.exit(56);
	 }

	 try
	 {
	    BufferedReader br = new BufferedReader( 
               new InputStreamReader(new FileInputStream(stackFile))); 
            while((line = br.readLine()) !=  null) 
            { 
               if(!line.startsWith("#")) 
               { 
                  if(line.indexOf(" ") != -1) 
                  { 
                     err("Spaces in stack name " + line); 
                     System.exit(21); 
                  } 
		  line = line.toLowerCase();
		  stacks.put(line, new Stack(line));
               } 
               lnum++; 
            } 
 
            br.close(); 
	 }
	 catch(FileNotFoundException f) 
         { 
            err("Can't find the unit listing file " + unitFile); 
            System.exit(55); 
         } 
         catch(IOException f) 
         { 
            err("Can't read from the unit listing file " + unitFile); 
            System.exit(56); 
         } 
      }

      public static void main(String[] args)
      {
	 boolean done = false;
	 boolean restored = false;

	 actionLog = new ArrayList();

	 String logfile = "logfile.txt";

	 printIntro();

	 if(args.length == 0)
	    getPlayers();
	 else
	 {
	    playerNames = args;
	    numPlayers = args.length+1;
	 }

	 for(int i = 0 ; i < playerNames.length ; i++)
	    playerNames[i] = playerNames[i].toLowerCase();


	 fillUnitTypesAndStacks("unitfile.txt", "stackfile.txt");

	 if( new File(logfile).exists() )
	 {
	    boolean decided = false;
	    
	    say("Log file " + logfile + " already exists.  You can choose a new");
	    say("  log file, restore from this one, overwrite this one, or quit:");
	    say("");
	    say("Select one: choose / restore / overwrite / quit");

	    while(!decided)
	    {
	       printPrompt();
	       String action = readLine();
	       if(action == null)
		  say("Try again");
	       else if(action.toLowerCase().equals("c") || action.toLowerCase().equals("choose"))
	       {
		  boolean chosen = false;
		  while(!chosen)
		  {
		     say("What file should we use for our logging?");
		     printPrompt();
		     logfile = readLine();
		     if(logfile != null)
			chosen = true;
		     else
			say("Try again");
		  }
		  decided = true;
	       }
	       else if(action.toLowerCase().equals("r") || action.toLowerCase().equals("restore"))
	       {
		  restored =true;
		  say("Restoring from log...");
		  try
		  {
		     readLog(logfile);
		  }
		  catch(ParserException pe)
		  {
		     say("Error in parsing log: " + pe.getErr());
		     say("Please fix and try again.");
		     System.exit(7);
		  }
		  decided = true;
	       }
	       else if(action.toLowerCase().equals("o") || action.toLowerCase().equals("overwrite"))
	       {
		  say("Overwriting the old log");
		  decided = true;
	       }
	       else if(action.toLowerCase().equals("q") || action.toLowerCase().equals("quit"))
	       {
		  say("Quitting...");
		  System.exit(0);
	       }
	       else
	       {
		  say("Not understood.  Please type one of choose / restore / overwrite / quit");
	       }
	    }
	 }

         logging = true;

	 try
	 {
	    // don't overwrite on restored games
	    logOut = new PrintWriter(new FileOutputStream(logfile, restored));
	    if(!restored)
	    {
	       logOut.print("# Game of ");
	       for(int i = 0 ; i < playerNames.length ; i++)
		  logOut.print(playerNames[i] + " ");
	       logOut.println("[" + System.currentTimeMillis() + "]");
	       logOut.flush();
	    }
	 }
	 catch(IOException ex)
	 {
	    err("Can't create log file.  Logging won't happen.");
	    logging = false;
	 }

	 while(!done)
	 {
	    printPrompt();
	    try
	    {
	       done = !processCommand(readLine());
	    }
	    catch(ParserException pe)
	    {
	       inputerr(pe.getErr());
	    }
	 }
      }

      public static int findPlayer(String name)
      {
	 for(int i = 0 ; i < playerNames.length ; i++)
	    if(playerNames[i].equals(name))
	       return i;
	 return -1;
      }

      public static boolean validType(int t)
      {
	 return t >= 0 && t < unitTypes.size();
      }

      public static boolean isValidStack(String s)
      {
	 return stacks.containsKey(s);
      }

      public static boolean isValidUnit(String u)
      {
	 return unitTypes.containsKey(u);
      }

      public static boolean isValidPlayer(String p)
      {
	 return p != null && findPlayer(p) != -1;
      }

      public static boolean isValidNumber(String s, int min, int max)
      {
	 try
	 {
	    int n = Integer.parseInt(s);
	    return n <= max && n >= min;
	 }
	 catch(NumberFormatException nfe)
	 {
	    return false;
	 }
      }

      public static void revealUnit(String stackStr, String unitStr, String playerStr, String numRevealed)
      {
	 Stack s = (Stack)stacks.get(stackStr);
//	 out("About to try unitTypes.get(unitStr)");
//	 out("  unitStr = " + unitStr);
//	 out("  unitTypes = " + unitTypes);
//	 out("  unitTypes.get(unitStr) = " + unitTypes.get(unitStr));
	 Integer typeI = (Integer)unitTypes.get(unitStr);
	 int p = -1;
	 int numR = -1;

/*
	 if(numRevealed != null)
	 {
	    try
	    {
	       numR = Integer.parseInt(numRevealed);
	    }
	    catch(NumberFormatException nfe)
	    {
	       inputerr("The number \"" + numRevealed + "\" was not properly interpreted, unit not revealed.");
	       return;
	    }
	 }
	 
	 if(s == null)
	 {
	    inputerr("Stack name " + stackStr + " not recognized, unit not revealed.");
	    return;
	 }
	 if(typeI == null || !validType(typeI.intValue()))
	 {
	    inputerr("Unit name " + unitStr + " not recognized, unit not revealed.");
	    return;
	 }
	 if(typeI == null || !validType(typeI.intValue()))
	 {
	    inputerr("Unit name " + unitStr + " not recognized, unit not revealed.");
	    return;
	 }
	 if(playerStr != null)
	 {
	    p = findPlayer(playerStr);
	    if(p == -1)
	    {
	       inputerr("Player name " + playerStr + " not recognized; unit not revealed.");
	       return;
	    }
	 }
*/
	 p = findPlayer(playerStr);
	 numR = Integer.parseInt(numRevealed);

	 if(numR == -1)
	    numR = 1;

/*	 if(numR < 1 || numR > 7)
	 {
	    inputerr("Revealing " + numR + " units of one type doesn't make sense.  Doing nothing.");
	    return;
	 }
*/

	 if(p == -1)
	    s.revealUnit(typeI.intValue(), numR);
	 else
	    s.revealUnit(typeI.intValue(), p, numR);
      }

      public static void addUnit(String stackStr, String unitStr)
      {
	 Stack s = (Stack)stacks.get(stackStr); 
	 /*       if(s == null) 
         { 
            inputerr("Stack name " + stackStr + " not recognized; unit not added."); 
            return; 
         } 
	 */
	 Integer typeI = (Integer)unitTypes.get(unitStr); 
/*	 if(typeI == null || !validType(typeI.intValue()))
	 { 
	    inputerr("Unit name " + unitStr + " not recognized; unit not added."); 
	    return; 
	 } 
*/
	 s.addUnit(new Unit(typeI.intValue(), s));
      }

      public static void splitStack(String oldStack, String newStack, int newSize)
      {
	 Stack oldS = (Stack)stacks.get(oldStack);  
/*
         if(oldS == null)  
         {  
            inputerr("Stack name " + oldStack + " not recognized; stack not split.");  
            return;  
         }  
*/
	 Stack newS = (Stack)stacks.get(newStack);  
/*
         if(newS == null)  
         {  
            inputerr("Stack name " + newStack + " not recognized; stack not split.");  
            return;  
         }  
*/
/*
	 if(newS.getSize() != 0)
	 {
	    inputerr("Stack " + newStack + " looks like its already in use; stack " + oldStack + " not split.");
	    return;
	 }
*/
/*	 if(oldS.getSize()-newSize < 2 || newSize < 2)
	 {
	    inputerr("You can't split a stack of size " + oldS.getSize() + " into stacks of");
	    inputerr("sizes " + (oldS.getSize()-newSize) + " and " + newSize + "; stacks not split.");
	    return;
	 }
*/
	 oldS.setSize(oldS.getSize()-newSize);
	 newS.setSize(newSize);

	 oldS.copyUnitlistOnto(newS);
      }

      public static void addStack(String stackStr)
      {
	 Stack s = new Stack(stackStr);
	 stacks.put(stackStr, s);
      }
      
      public static void deleteStack(String stackStr)
      {
	 Stack s = (Stack)stacks.get(stackStr);  
/*         if(s == null)  
         {  
            inputerr("Stack name " + stackStr + " not recognized; stack not deleted.");  
            return;  
         }  
*/
	 s.setSize(0);
	 /* maybe need to tell s to remove possibilities */
      }

      /* note : units are only deleted when the whole stack is known, so it doesn't
	 matter which one we delete.  To be safe, though, we'll follow the procedure 
	 for reveal.*/

      public static void deleteUnit(String stackStr, String unitStr)
      {
	 Stack s = (Stack)stacks.get(stackStr);   
/*
         if(s == null)   
         {   
            inputerr("Stack name " + stackStr + " not recognized; unit not deleted.");   
            return;   
         }   
*/
         Integer typeI = (Integer)unitTypes.get(unitStr); 
/*         if(typeI == null || !validType(typeI.intValue())) 
         { 
            inputerr("Unit name " + unitStr + " not recognized, unit not revealed."); 
            return; 
         } 
*/
	 s.removeUnit(typeI.intValue());
      }

      public static void doUndo()
      {
	 out("undo not implemented.  To get undo-like functionality, quit, and edit");
	 out("  the logfile (probably logfile.txt).  Start again and choose \"restore\".");
      }

      public static void listUnitTypes()
      {
	 Iterator it = unitTypes.keySet().iterator();
	 while(it.hasNext())
	    out((String)it.next());
      }

      public static void listPlayers()
      {
	 for(int i = 0 ; i < playerNames.length ; i++)
	    out(playerNames[i]);
      }

      public static void listStacks(boolean onlyActive)
      {
	 Iterator it = (stacks.values()).iterator();
	 Stack s;
	 while(it.hasNext())
	 {
	    s = (Stack)it.next();
	    if(s.getSize() > 0 || !onlyActive)
	       out(s.getName());
	 }	       
      }

      public static void viewLog()
      {
	 Iterator it = actionLog.iterator();
	 int num = 0;
	 while(it.hasNext())
	 {
	    out(num + " " + it.next());
	    num++;
	 }
      }

      public static void viewStack(String stackStr, String playerStr)
      {
	 Stack s = (Stack)stacks.get(stackStr);   
/*         if(s == null)   
         {   
            inputerr("Stack name " + stackStr + " not recognized; stack not shown.");   
            return;   
         }   
*/
	 int p = -1;
	 if(playerStr != null)
	 {
	    p = findPlayer(playerStr); 
/*            if(p == -1) 
            { 
               inputerr("Player name " + playerStr + " not recognized; stack not shown."); 
               return; 
	    }
*/
	 }
	 if(p == -1)
	    s.view();
	 else
	    s.view(p);
      }


      public static void out(String s)
      {
	 say(s);
      }
      public static void say(String s)
      {
	 System.out.println(s);
      }

      public static void inputerr(String s)
      {
	 say("parser err: " + s);
      }
      
      public static void err(String s)
      {
	 System.err.println("Titan: " + s);
      }


}